/***************************************************************************
OctreeCHCscenemanager.h  -  description
-------------------
begin                : Fri Sep 27 2002
copyright            : (C) 2002 by Jon Anderson
email                : janders@users.sf.net
***************************************************************************/

/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#ifndef __OctreeCHCSCENEMANAGER_H
#define __OctreeCHCSCENEMANAGER_H

#include "OgreSceneManager.h"
#include "OgreRenderOperation.h"
#include "OgreSphere.h"

#include <list>
#include <algorithm>

#include "OctreeCHCPrerequisites.h"

#include "OctreeCHC.h"
#include "OctreeCHCOcclusion.h"
#include "OgreRenderable.h"

#include "OctreeCHCPoolSet.h"
#include "OgreSpatialTreeManager.h"
#include "OgreSpatialTree.h"
#include "OgreSpatialTreeNode.h"

namespace OgreOctreeCHC
{

    /// Factory for OctreeSceneManager
    class OctreeCHCSceneManagerFactory : public Ogre::SceneManagerFactory
    {
    protected:
        void initMetaData(void) const;
    public:
        OctreeCHCSceneManagerFactory() {}
        ~OctreeCHCSceneManagerFactory() {}
        /// Factory type name
        static const Ogre::String FACTORY_TYPE_NAME;
        Ogre::SceneManager* createInstance(const Ogre::String& instanceName);
        void destroyInstance(Ogre::SceneManager* instance);
    };

	/// Identification struct that permits temporal coherence between renders.
	struct RenderTemporalIdentification
	{
		Ogre::Viewport *mViewport;
		Ogre::RenderTarget *mTarget;
		Ogre::Camera *mCamera;
		bool mOnlyShadowCasters;
		unsigned int id;
		bool mRegistered;
		culling_modes mCullMode;
		unsigned int mLastFrameId;
	};

    class OctreeCHCNode;

    class OctreeCHCCamera;
    class OctreeCHCIntersectionSceneQuery;
    class OctreeCHCRaySceneQuery;
    class OctreeCHCSphereSceneQuery;
    class OctreeCHCAxisAlignedBoxSceneQuery;
    class OctreeCHCPlaneBoundedVolumeListSceneQuery;

    /** Specialized SceneManager that divides the geometry into an OctreeCHC in order to facilitate spatial queries.
    @remarks
    For debugging purposes, a special "CullCamera" can be defined.  To use it, call setUseCallCamera(true),
    and create a camera named "CullCamera".  All culling will be performed using that camera, instead of the viewport
    camera, allowing you to fly around and examine culling.
    */

    class OctreeCHCSceneManager : public Ogre::SceneManager
    {
        friend class OctreeCHCIntersectionSceneQuery;
        friend class OctreeCHCRaySceneQuery;
        friend class OctreeCHCSphereSceneQuery;
        friend class OctreeCHCAxisAlignedBoxSceneQuery;
        friend class OctreeCHCPlaneBoundedVolumeListSceneQuery;

    public:
		enum SpatialTreeType
		{
			LOOSE_OCTREE = 0
			,SPHERE_OCTREE
			,ABT_OCTREE
			,NUM_SPATIAL_TREE
		};

    public:
        static int intersect_call;
        /** Standard Constructor.  Initializes the OctreeCHC to -500,-500,-500 to 500,500,500 with unlimited depth. */
        OctreeCHCSceneManager(const Ogre::String &name);
        /** Standard Constructor */
        OctreeCHCSceneManager(const Ogre::String &name, Ogre::AxisAlignedBox& box, int max_depth);
        /** Standard destructor */
        virtual ~OctreeCHCSceneManager(void);
        
        void shutdown();
        /// @copydoc SceneManager::getTypeName
        const Ogre::String& getTypeName(void) const;

        /** Initializes the manager to the given box and depth.
        */
        void init(const Ogre::AxisAlignedBox& box, int d);

        /** Creates a specialized OctreeCHCNode */
        virtual	Ogre::SceneNode* createSceneNode(void);
        /** Creates a specialized OctreeCHCNode */
        virtual Ogre::SceneNode* createSceneNode(const Ogre::String& name);
        /** Creates a specialized OctreeCHCCamera */
        virtual Ogre::Camera* createCamera(const Ogre::String& name);
        virtual void destroyCamera(Ogre::Camera *cam);
        virtual void destroyCamera(const Ogre::String& name);
        virtual void destroyAllCameras(void);

        void addCamera(Ogre::Camera *cam);

        /** Deletes a scene node */
        virtual void destroySceneNode(const Ogre::String& name);

		/** Recurses through the OctreeCHC determining which nodes are visible. */
        virtual void _findVisibleObjects(Ogre::Camera* cam, Ogre::VisibleObjectsBoundsInfo *visibleBounds, bool onlyShadowCasters);


        /** Alerts each un-culled object, notifying it that it will be drawn.
        * Useful for doing calculations only on nodes that will be drawn, prior
        * to drawing them...
        */
        //virtual void _alertVisibleObjects(void);

        /** Walks through the OctreeCHC, adding any visible objects to the render queue.
        @remarks
	        If any octant in the OctreeCHC if completely within the the view frustum,
	        all sub-children are automatically added with no visibility tests.
        */
        
        /** Checks the given OctreeCHCNode, and determines if it needs to be moved
        * to a different octant.
        */
        void _updateSceneNode(Ogre::SceneNode* node);
        
	    /** Removes the given OctreeCHC node */
        void _removeSceneNode(Ogre::SceneNode* node);
        
	    /** Adds the OctreeCHC Node, starting at the given OctreeCHC, and recursing at max to the specified depth.
        */
        void _addSceneNode(Ogre::SceneNode* ocnod, SpatialTree *tree, const int depth = 0);

	    /** Adds the OctreeCHC Node, starting at the given OctreeCHC, and recursing at max to the specified depth.
        */
        void _addMovableSceneNode(Ogre::SceneNode* ocnod, SpatialTree *tree, int depth = 0);

	    /** Adds the OctreeCHC Node, starting at the given OctreeCHC, and recursing at max to the specified depth.
        */
        void _addStaticSceneNode(Ogre::SceneNode* ocnod, SpatialTree *tree, int depth = 0);

        /** Sets the box visibility flag */
        void setShowBoxes(bool b)
        {
             mShowBoxes = b;
        };

        /** Sets the cull camera flag */
        void setUseCullCamera(bool b)
        {            
             mCullCamera = b;   
        };

		/** Sets the cull debug flag */
		void setCullDebug(bool b)
		{		
			mCullDebug = b;	
		};

        /** Sets the given option for the SceneManager
                @remarks
            Options are:
            "Size", Ogre::AxisAlignedBox *;
            "CullCamera", bool *;
            "Depth", int *;
            "ShowOctree", bool *;
        */

        virtual bool setOption(const Ogre::String& key, const void* value);

        /** Gets the given option for the Scene Manager.
            @remarks
            See setOption
        */
        virtual bool getOption(const Ogre::String& key, void* value);

        bool getOptionValues(const Ogre::String& key, Ogre::StringVector& refValueList);
        bool getOptionKeys(Ogre::StringVector& refKeys);
        /** Overridden from SceneManager */
        void clearScene(void);

        Ogre::AxisAlignedBoxSceneQuery* createAABBQuery(const Ogre::AxisAlignedBox& box, unsigned long mask);
        Ogre::SphereSceneQuery* createSphereQuery(const Ogre::Sphere& sphere, unsigned long mask);
        Ogre::PlaneBoundedVolumeListSceneQuery* createPlaneBoundedVolumeQuery(const Ogre::PlaneBoundedVolumeList& volumes, unsigned long mask);
        Ogre::RaySceneQuery* createRayQuery(const Ogre::Ray& ray, unsigned long mask);
        Ogre::IntersectionSceneQuery* createIntersectionQuery(unsigned long mask);

        void directRenderSingleQueue(Ogre::RenderQueue *queue);
        void directRenderSingleObject(Ogre::Renderable *mo);

        void addVisible(Ogre::MovableObject *mo);

        void registeredNodeInCamera(OcclusionElement *on);
		void unregisteredNodeInCamera(OcclusionElement *on);

		void setDebugText(const Ogre::String &t){mDebugText = t;}
		const Ogre::String &getDebugText() const {return mDebugText;}

		OctreeCHCCamera* getCurrentCurrentOptionCamera();
		void setCurrentCurrentOptionCamera(OctreeCHCCamera*);


		/** Prompts the class to send its contents to the renderer.
		@remarks
		This method prompts the scene manager to send the
		contents of the scene it manages to the rendering
		pipeline, possibly preceded by some sorting, culling
		or other scene management tasks. Note that this method is not normally called
		directly by the user application; it is called automatically
		by the Ogre rendering loop.
		@param camera Pointer to a camera from whose viewpoint the scene is to
		be rendered.
		@param vp The target viewport
		@param includeOverlays Whether or not overlay objects should be rendered
		*/
		virtual void _renderScene(Ogre::Camera* camera, Ogre::Viewport* vp, bool includeOverlays);



		/////////////////////////////////
		/// TreeManager Related
        /** Resizes the OctreeCHC to the given size */
		void resize(const Ogre::AxisAlignedBox& box);
        void resize(const Ogre::AxisAlignedBox &box, const int depth);

		// mMaxDepth getter
		int getMaxDepth() const;
		// mMaxDepth setter
		void setMaxDepth(int val);

        const Ogre::AxisAlignedBox &getBoundingBox()const;

		SpatialTreeManager* getTreeManager(){return mTreeManager;}
    protected:

        MovableObjectList * mCamInProgressVisibles;
        Ogre::PolygonMode mBeforeHWOCCamDetail;
        bool mBeforeHWOCCamClear;

        // OCCLUSION 
        OcclusionCulling mCulling;

        //DEBUG  INFO
        
            /// box visibility flag
            bool mShowBoxes;
            /// list of boxes to be rendered
            MovableObjectList mVisualDebugBoundingBoxes;
            /// cull camera flag
            bool mCullCamera;
            /// Overlay Culling visual info flag
			bool mCullDebug;
			/// Overlay Culling text info flag
			bool mCullDebugText;

        

		Ogre::String mDebugText;

		OctreeCHCCamera *mCurrentOptionCamera;

        Ogre::RenderTexture     *mOcclusionDepth;
        Ogre::Camera            *mOcclusionCamera;

        void enableHardwareOcclusionTests();
        void disableHardwareOcclusionTests();

        SpatialTreeManager *mTreeManager;

		SpatialTreeType mCurrentSpatialManagerType;


		void removeRenderTemporalIdentification(OctreeCHCCamera * cam);
		bool setRenderTemporalIdentification(OctreeCHCCamera * const cam, const bool onlyShadowCasters);
		std::vector <RenderTemporalIdentification *> mRenderTemporalIdentifications;
		unsigned int mRenderTemporalIdentificationIdCount;
    };


}

#endif //__OctreeCHCSCENEMANAGER_H

