/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/
/***************************************************************************
OctreeCHCnode.h  -  description
-------------------
begin                : Fri Sep 27 2002
copyright            : (C) 2002 by Jon Anderson
email                : janders@users.sf.net

Enhancements 2003 - 2004 (C) The OGRE Team
***************************************************************************/
#ifndef __OgreSpatialTreeNODE_H
#define __OgreSpatialTreeNODE_H

#include "OctreeCHCPrerequisites.h"

#include "OgreSceneNode.h"
#include "OctreeCHCOcclusionElement.h"

namespace OgreOctreeCHC
{

    /** Specialized SceneNode that is customized for working within an OctreeCHC. Each node
    * maintains it's own bounding box, rather than merging it with all the children.
    *
    */
	class SpatialTreeNode : public Ogre::SceneNode, public OcclusionElement
    {
	public:
		//-----------------------------------------------------------------------
		/** Standard constructor */
		SpatialTreeNode(Ogre::SceneManager* creator);
		  //-----------------------------------------------------------------------
		  SpatialTreeNode::SpatialTreeNode(Ogre::SceneManager* creator, const Ogre::String& name) ;
		/** Standard destructor */
        ~SpatialTreeNode(void);

        /** Returns the Tree in which this OctreeCHCNode resides
        */
        SpatialTree* getTreeParent(void) ;
        /** Sets the Tree in which this OctreeCHCNode resides
        */
        void setTreeParent(SpatialTree* t);

        /** Adds all the attached scene nodes to the render queue
        */
		virtual void _findVisibleObjects
			(
			Ogre::Camera* cam, 
			Ogre::RenderQueue* queue, 
			Ogre::VisibleObjectsBoundsInfo* visibleBounds, 
			bool includeChildren, 
			bool displayNodes, 
			bool onlyShadowCasters
			);
        
        MovableObjectList *getVisibleNotifiedNodeObjects(Ogre::Camera*  cam, 
                                    const bool onlyShadowCasters);
         
        bool notifyNodeObjects(Ogre::Camera*  cam, 
                                const bool onlyShadowCasters);
        /** Adds all the attached scene nodes to the render queue
        * with the particularity that it has already been notified by camera.
        */
        void _addAlreadyNotifiedToVisibles();
        void _addToVisibles(Ogre::Camera*  cam, const bool onlyShadowCasters);
        /** Sets up the LegacyRenderOperation for rendering this scene node as geometry.
        @remarks
        This will render the scene node as a bounding box.
        */
        //virtual void getRenderOperation(RenderOperation& op);

      
        /** Adds a (pre-created) child scene node to this node. If it is attached to another node,
            it must be detached first.
        @param child The Node which is to become a child node of this one
        */
        virtual void addChild(Ogre::Node* child);
        /** Drops the specified child from this node. 
        @remarks
            Does not delete the node, just detaches it from
            this parent, potentially to be reattached elsewhere. 
            There is also an alternate version which drops a named
            child from this node.
        */
        virtual Node* removeChild(unsigned short index);
        /** Drops the specified child from this node. 
        @remarks
        Does not delete the node, just detaches it from
        this parent, potentially to be reattached elsewhere. 
        There is also an alternate version which drops a named
        child from this node.
        */
        virtual Node* removeChild(Ogre::Node* child);

        /** Drops the named child from this node. 
        @remarks
            Does not delete the node, just detaches it from
            this parent, potentially to be reattached elsewhere.
        */
        virtual Node* removeChild(const Ogre::String& name);
        /** Removes all child Nodes attached to this node. Does not delete the nodes, just detaches them from
            this parent, potentially to be reattached elsewhere.
        */
        virtual void removeAllChildren(void);

        
        inline bool isLeaf() const;

        inline void depthTraversal(DepthTraversal&tr);
        inline void depthTraversal(const DepthTraversalConst &tr);
        inline void depthTraversal(const ConstDepthTraversalConst &tr) const;
		inline void breadthTraversal(BreadthTraversal&tr);
		inline void breadthTraversal(const BreadthTraversalConst &tr);
		inline void breadthTraversal(const ConstBreadthTraversalConst &tr) const;

        
        void setDebugCorners();
            

         /** Adds an instance of a scene object to this node.
            @remarks
                Scene objects can include Entity objects, Camera objects, Light objects, 
                ParticleSystem objects etc. Anything that subclasses from MovableObject.
        */
        virtual void attachObject(Ogre::MovableObject* obj);   
        /** Detaches the indexed object from this scene node.
        @remarks
            Detaches by index, see the alternate version to detach by name. Object indexes
            may change as other objects are added / removed.
        */
        virtual Ogre::MovableObject* detachObject(unsigned short index);
        /** Detaches an object by pointer. */
        virtual void detachObject(Ogre::MovableObject* obj);

        /** Detaches the named object from this node and returns a pointer to it. */
        virtual Ogre::MovableObject* detachObject(const Ogre::String& name);

        /** Detaches all objects attached to this node.
        */
        virtual void detachAllObjects(void);

        inline bool isOccluder () const;

        bool isStaticNode() const;

        void setStaticCulling (const bool staticNode);


    protected:
        /** Internal method for updating the bounds for this OctreeCHCNode.
        @remarks
			This method determines the bounds solely from the attached objects, not
			any children. If the node has changed it's bounds, it is removed from its
			current OctreeCHC, and reinserted into the tree.
        */
        void _updateBounds(void);

        void _removeNodeAndChildren(void);

        ///Tree this node is attached to.
        SpatialTree*					mTreeParent;
        /// if node need to in loose octree or normal octree
        bool                    mIsStaticNode;
        /// if node need to be drawn during query phase
        bool                    mIsOccluder;
        // if a pre-notified object is visible.
        MovableObjectList       mNotifiedVisibles;
            
    };
}

#include "OctreeCHCSceneManager.h"
#include "OctreeCHC.h"

namespace OgreOctreeCHC
{
	//-----------------------------------------------------------------------
	//-----------------------------------------------------------------------
	//-----------------------------------------------------------------------
	inline SpatialTree* SpatialTreeNode::getTreeParent(void) 
	{
		return mTreeParent;
	};
	//-----------------------------------------------------------------------
	inline void SpatialTreeNode::setTreeParent(SpatialTree* parent)
	{
		mTreeParent = parent;
		OcclusionElement::setOcclusionParent(parent);
	};
	//-----------------------------------------------------------------------
	inline bool SpatialTreeNode::isLeaf() const {return true;};
	//-----------------------------------------------------------------------
	inline void SpatialTreeNode::depthTraversal(DepthTraversal&tr)
	{   
		tr.onDepthLeaf (*this);
	};
	//-----------------------------------------------------------------------
	inline void SpatialTreeNode::depthTraversal(const DepthTraversalConst &tr)
	{  
		tr.onDepthLeaf(*this);
	};
	//-----------------------------------------------------------------------
	inline void SpatialTreeNode::depthTraversal(const ConstDepthTraversalConst &tr) const
	{   
		tr.onDepthLeaf (*this);
	};         
	//-----------------------------------------------------------------------  
	inline void SpatialTreeNode::breadthTraversal(BreadthTraversal&tr)
	{   
		tr.onBreadthLeaf (*this);
	};
	//-----------------------------------------------------------------------
	inline void SpatialTreeNode::breadthTraversal(const BreadthTraversalConst &tr)
	{  
		tr.onBreadthLeaf(*this);
	};
	//-----------------------------------------------------------------------
	inline void SpatialTreeNode::breadthTraversal(const ConstBreadthTraversalConst &tr) const
	{   
		tr.onBreadthLeaf (*this);
	};  
	//-----------------------------------------------------------------------
	inline bool SpatialTreeNode::isOccluder () const  { return mIsOccluder; }    
	//-----------------------------------------------------------------------
	inline bool SpatialTreeNode::isStaticNode() const { return mIsStaticNode; }
	//-----------------------------------------------------------------------
	inline void SpatialTreeNode::setStaticCulling (const bool staticNode) { mIsStaticNode = staticNode; };           
	//-----------------------------------------------------------------------
}


#endif //__OgreSpatialTreNODE_H
