
#include "OctreeCHCPrecompiledHeaders.h"

#include "OctreeCHCQueryTraversal.h"

using namespace Ogre;

namespace OgreOctreeCHC
{	 	
	//---------------------------------------------------------------------
	enum Intersection
	{
		OUTSIDE		=	0,
		INSIDE		=	1,
		INTERSECT	=	2
	};
	//---------------------------------------------------------------------
	Intersection intersect(const Ray &ray, const Ogre::AxisAlignedBox &box)
	{
		// Null box?
		if (box.isNull()) 
			return OUTSIDE;
		// Infinite box?
		if (box.isInfinite()) 
			return INTERSECT;

		const Ogre::Vector3& min = box.getMinimum();
		const Ogre::Vector3& max = box.getMaximum();
		const Ogre::Vector3& rayorig = ray.getOrigin();

		// Check origin inside first
		if ( rayorig > min && rayorig < max )
			return INSIDE;

		const Ogre::Vector3& raydir = ray.getDirection();

		Real lowt = 0.0f;
		Real t;
		bool hit = false;
		Ogre::Vector3 hitpoint;

		// Check each face in turn, only check closest 3
		// Min x
		if (rayorig.x < min.x && raydir.x > 0)
		{
			t = (min.x - rayorig.x) / raydir.x;
			if (t > 0)
			{
				// Substitute t back into ray and check bounds and dist
				hitpoint = rayorig + raydir * t;
				if (hitpoint.y >= min.y && hitpoint.y <= max.y &&
					hitpoint.z >= min.z && hitpoint.z <= max.z &&
					(!hit || t < lowt))
				{
					return INTERSECT;
				}
			}
		}
		// Max x
		if (rayorig.x > max.x && raydir.x < 0)
		{
			t = (max.x - rayorig.x) / raydir.x;
			if (t > 0)
			{
				// Substitute t back into ray and check bounds and dist
				hitpoint = rayorig + raydir * t;
				if (hitpoint.y >= min.y && hitpoint.y <= max.y &&
					hitpoint.z >= min.z && hitpoint.z <= max.z &&
					(!hit || t < lowt))
				{
					return INTERSECT;
				}
			}
		}
		// Min y
		if (rayorig.y < min.y && raydir.y > 0)
		{
			t = (min.y - rayorig.y) / raydir.y;
			if (t > 0)
			{
				// Substitute t back into ray and check bounds and dist
				hitpoint = rayorig + raydir * t;
				if (hitpoint.x >= min.x && hitpoint.x <= max.x &&
					hitpoint.z >= min.z && hitpoint.z <= max.z &&
					(!hit || t < lowt))
				{
					return INTERSECT;
				}
			}
		}
		// Max y
		if (rayorig.y > max.y && raydir.y < 0)
		{
			t = (max.y - rayorig.y) / raydir.y;
			if (t > 0)
			{
				// Substitute t back into ray and check bounds and dist
				hitpoint = rayorig + raydir * t;
				if (hitpoint.x >= min.x && hitpoint.x <= max.x &&
					hitpoint.z >= min.z && hitpoint.z <= max.z &&
					(!hit || t < lowt))
				{
					return INTERSECT;
				}
			}
		}
		// Min z
		if (rayorig.z < min.z && raydir.z > 0)
		{
			t = (min.z - rayorig.z) / raydir.z;
			if (t > 0)
			{
				// Substitute t back into ray and check bounds and dist
				hitpoint = rayorig + raydir * t;
				if (hitpoint.x >= min.x && hitpoint.x <= max.x &&
					hitpoint.y >= min.y && hitpoint.y <= max.y &&
					(!hit || t < lowt))
				{
					return INTERSECT;
				}
			}
		}
		// Max z
		if (rayorig.z > max.z && raydir.z < 0)
		{
			t = (max.z - rayorig.z) / raydir.z;
			if (t > 0)
			{
				// Substitute t back into ray and check bounds and dist
				hitpoint = rayorig + raydir * t;
				if (hitpoint.x >= min.x && hitpoint.x <= max.x &&
					hitpoint.y >= min.y && hitpoint.y <= max.y &&
					(!hit || t < lowt))
				{
					return INTERSECT;
				}
			}
		}
		return OUTSIDE;
	}

	//---------------------------------------------------------------------
	Intersection intersect2(const Ray &one, const Ogre::AxisAlignedBox &two)
	{
		OctreeCHCSceneManager::intersect_call++;

		// Null box?
		if (two.isNull()) 
			return OUTSIDE;
		// Infinite box?
		if (two.isInfinite()) 
			return INTERSECT;

		const Ogre::Vector3* const pCorners = two.getAllCorners();
		const Ogre::Vector3 origin = one.getOrigin();
		const Ogre::Vector3 dir = one.getDirection();

		Ogre::Vector3 maxT (-1.0f, -1.0f, -1.0f);

		unsigned int i = 0;
		bool inside = true;
		for (i = 0; i < 3; i++)
		{
			if (origin[i] < pCorners[0][i])
			{
				inside = false;
				if(dir[i] > 0)
					maxT[i] = (pCorners[0][i] - origin[i]) / dir[i];
			}
			else if (origin[i] > pCorners[4][i])
			{
				inside = false;
				if(dir[i] < 0)
					maxT[i] = (pCorners[4][i] - origin[i]) / dir[i];
			}
		}

		if(inside)
		{
			return INTERSECT;
		}
		int whichPlane = 0;
		if(maxT[1] > maxT[whichPlane])
			whichPlane = 1;
		if(maxT[2] > maxT[whichPlane])
			whichPlane = 2;

		if(((int)maxT[whichPlane]) & 0x80000000)
		{
			return OUTSIDE;
		}
		const Real maxTThisPlane = maxT[whichPlane];
		for(i = 0; i < 3; i++)
		{
			if(i != whichPlane)
			{
				const Real f = origin[i] + maxTThisPlane * dir[i];
				if (f < (pCorners[0][i] - 0.00001f) ||
					f > (pCorners[4][i] + 0.00001f))
				{
					return OUTSIDE;
				}
			}
		}

		return INTERSECT;

	}
	//---------------------------------------------------------------------
	/** Checks how the second box intersects with the first.
	*/
	Intersection intersect(const PlaneBoundedVolume &one, const Ogre::AxisAlignedBox &two)
	{
		OctreeCHCSceneManager::intersect_call++;
		// Null box?
		if (two.isNull()) 
			return OUTSIDE;
		if (two.isInfinite()) 
			return INTERSECT;

		// Get corners of the box
		const Ogre::Vector3 * const pCorners = two.getAllCorners();

		// For each plane, see if all points are on the negative side
		// If so, object is not visible.
		// If one or more are, it's partial.
		// If all aren't, full
		static const int corners[ 8 ] = {0, 4, 3, 5, 2, 6, 1, 7};

		bool all_inside = true;
		PlaneList::const_iterator i, iend = one.planes.end();
		for (i = one.planes.begin(); i != iend; ++i)
		{
			const Plane& plane = *i;
			bool all_outside = true;
			for (unsigned int corner = 0; corner < 8; ++corner)
			{
				const Real distance = plane.getDistance(pCorners[ corners[ corner ] ]);
				all_outside = all_outside && (distance < 0);
				all_inside = all_inside && (distance >= 0);

				if (!all_outside && !all_inside)
					break;
			}

			if (all_outside)
				return OUTSIDE;
		}

		if (all_inside)
			return INSIDE;
		else
			return INTERSECT;

	}
	//---------------------------------------------------------------------
	/** Checks how the second box intersects with the first.
	*/
	Intersection intersect(const Ogre::AxisAlignedBox &one, const Ogre::AxisAlignedBox &two)
	{
		OctreeCHCSceneManager::intersect_call++;

		// Null box?
		if (one.isNull() || two.isNull()) 
			return OUTSIDE;
		if (one.isInfinite()) 
			return INSIDE;
		if (two.isInfinite()) 
			return INTERSECT;

		const Ogre::Vector3 * outside = one.getAllCorners();
		const Ogre::Vector3 *inside = two.getAllCorners();

		if (inside[ 4 ].x < outside[ 0 ].x ||
			inside[ 4 ].y < outside[ 0 ].y ||
			inside[ 4 ].z < outside[ 0 ].z ||
			inside[ 0 ].x > outside[ 4 ].x ||
			inside[ 0 ].y > outside[ 4 ].y ||
			inside[ 0 ].z > outside[ 4 ].z)
		{
			return OUTSIDE;
		}

		const bool full = (inside[ 0 ].x > outside[ 0 ].x &&
			inside[ 0 ].y > outside[ 0 ].y &&
			inside[ 0 ].z > outside[ 0 ].z &&
			inside[ 4 ].x < outside[ 4 ].x &&
			inside[ 4 ].y < outside[ 4 ].y &&
			inside[ 4 ].z < outside[ 4 ].z);

		if (full)
			return INSIDE;
		else
			return INTERSECT;

	}
	//---------------------------------------------------------------------
	/** Checks how the box intersects with the sphere.
	*/
	Intersection intersect(const Sphere &one, const Ogre::AxisAlignedBox &two)
	{
		OctreeCHCSceneManager::intersect_call++;

		// Null box?
		if (two.isNull()) 
			return OUTSIDE;
		if (two.isInfinite()) 
			return INTERSECT;


		const Real sradius = Math::Sqr (one.getRadius());
		const Ogre::Vector3 scenter = one.getCenter();
		const Ogre::Vector3 * const corners = two.getAllCorners();

		const Ogre::Vector3 mndistance = (corners[ 0 ] - scenter);
		const Ogre::Vector3 mxdistance = (corners[ 4 ] - scenter);


		if (mndistance.squaredLength() < sradius &&
			mxdistance.squaredLength() < sradius)
		{
			return INSIDE;
		}

		//find the square of the distance
		//from the sphere to the box
		Real d = 0;
		for (unsigned int i = 0 ; i < 3 ; i++)
		{
			const Real sCenteri = scenter[i];
			if (sCenteri < corners[ 0 ][i])
			{
				const Real s = sCenteri - corners[ 0 ][i];
				d += s * s;
			}
			else if (sCenteri > corners[ 4 ][i])
			{
				const Real s = sCenteri - corners[ 4 ][i];
				d += s * s;
			}

		}

		const bool partial = (d <= sradius);

		if (!partial)
		{
			return OUTSIDE;
		}

		else
		{
			return INTERSECT;
		}


	}

	//-----------------------------------------------------------------------
	void SphereQueryDepthTraversal::onDepthTree(SpatialTree &node)
	{
		bool isFull = mIsVisible;
		if (!mIsVisible)
		{
			const Intersection isect = intersect(mSphere, node._getCullBounds());
			if (isect == OUTSIDE)
				return ;
			mIsVisible = (isect == INSIDE);
		}

		// Has to restore Parent Vis so that bothers can gets correct parent vis.
		const bool oldVis = mIsVisible;
		depthTraverseChildren (node);
		mIsVisible = oldVis; 
	}
	//-----------------------------------------------------------------------
	void SphereQueryDepthTraversal::onDepthLeaf (SpatialTreeNode &node)
	{
		if (&node != mExclude)
		{
			if (mIsVisible)
			{
				mList.push_back(&node);
			}
			else
			{
				const Intersection nsect = intersect(mSphere, 
					node._getWorldAABB());
				if (nsect != OUTSIDE)
				{
					mList.push_back(&node);
				}
			}
		}
	}
	//-----------------------------------------------------------------------
	//-----------------------------------------------------------------------
	//-----------------------------------------------------------------------
	void PlaneBoundedVolumeQueryDepthTraversal::onDepthTree(SpatialTree &node)
	{
		bool isFull = mIsVisible;
		if (!mIsVisible)
		{
			const Intersection isect = intersect(mPlaneBoundedVolume, node._getCullBounds());

			if (isect == OUTSIDE)
				return ;

			mIsVisible = (isect == INSIDE);
		}

		// Has to restore Parent Vis so that bothers can gets correct parent vis.
		const bool oldVis = mIsVisible;
		depthTraverseChildren (node);
		mIsVisible = oldVis; 
	}
	//-----------------------------------------------------------------------
	void PlaneBoundedVolumeQueryDepthTraversal::onDepthLeaf (SpatialTreeNode &node)
	{
		if (&node != mExclude)
		{
			if (mIsVisible)
			{
				mList.push_back(&node);
			}
			else
			{
				const Intersection nsect = intersect(mPlaneBoundedVolume, 
					node._getWorldAABB());
				if (nsect != OUTSIDE)
				{
					mList.push_back(&node);
				}
			}
		}
	}
	//-----------------------------------------------------------------------
	//-----------------------------------------------------------------------
	//-----------------------------------------------------------------------
	void RayQueryDepthTraversal::onDepthTree(SpatialTree &node)
	{
		bool isFull = mIsVisible;
		if (!mIsVisible)
		{
			const Intersection isect = intersect(mRay, node._getCullBounds());
			if (isect == OUTSIDE)
				return ;
			mIsVisible = (isect == INSIDE);
		}
		// Has to restore Parent Vis so that bothers can gets correct parent vis.
		const bool oldVis = mIsVisible;
		depthTraverseChildren (node);
		mIsVisible = oldVis; 
	}
	//-----------------------------------------------------------------------
	void RayQueryDepthTraversal::onDepthLeaf (SpatialTreeNode &node)
	{
		if (&node != mExclude)
		{
			if (mIsVisible)
			{
				mList.push_back(&node);
			}
			else
			{
				const Intersection nsect = intersect(mRay, 
					node._getWorldAABB());
				if (nsect != OUTSIDE)
				{
					mList.push_back(&node);
				}
			}
		}
	}
	//-----------------------------------------------------------------------
	//-----------------------------------------------------------------------
	//-----------------------------------------------------------------------
	void BoxQueryDepthTraversal::onDepthTree(SpatialTree &node)
	{
		bool isFull = mIsVisible;
		if (!mIsVisible)
		{
			const Intersection isect = intersect(mBoundingBox, node._getCullBounds());
			if (isect == OUTSIDE)
				return ;
			mIsVisible = (isect == INSIDE);
		}

		// Has to restore Parent Vis so that bothers can gets correct parent vis.
		const bool oldVis = mIsVisible;
		depthTraverseChildren (node);
		mIsVisible = oldVis; 
	}
	//-----------------------------------------------------------------------
	void BoxQueryDepthTraversal::onDepthLeaf (SpatialTreeNode &node)
	{
		if (&node != mExclude)
		{
			if (mIsVisible)
			{
				mList.push_back(&node);
			}
			else
			{
				const Intersection nsect = intersect(mBoundingBox, 
					node._getWorldAABB());
				if (nsect != OUTSIDE)
				{
					mList.push_back(&node);
				}
			}
		}
	}
	//-----------------------------------------------------------------------
}

