/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later__add
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/
/***************************************************************************
Tree.cpp  -  description
-------------------
begin                : Mon Sep 30 2002
copyright            : (C) 2002 by Jon Anderson
email                : janders@users.sf.net

Enhancements 2003 - 2004 (C) The OGRE Team

***************************************************************************/

#include "OctreeCHCPrecompiledHeaders.h"

#include "OgreWireBoundingBox.h"
#include "OctreeCHCOcclusionBoundingBox.h"

#include "OctreeCHCSceneManager.h"

#include "OgreSpatialTreeManager.h"
#include "OgreSpatialTree.h"
#include "OgreSpatialTreeNode.h"

#include "OgreStringConverter.h"

using namespace Ogre;

namespace OgreOctreeCHC
{
	//-----------------------------------------------------------------------
	SpatialTree::SpatialTree(SpatialTreeManager *treeManager) : 
		OcclusionElement(),
		mChildren(0), 
		mNumNodes(0),
		mTreeManager(treeManager),
		mCullBoundingBox(0),
		mWireBoundingBox(0),
		mTreeParent(0)
	{
	}
	//-----------------------------------------------------------------------
	SpatialTree::~SpatialTree(void)
	{		
		if (mChildren)
		{
			//initialize all children to null.
#ifdef _DEBUG
			for (unsigned int i = 0; i <  mTreeManager->getMaxChildrenCount(); i++)
			{
				octree_assert (mChildren[i] == 0);          
			}
#endif //_DEBUG
			delete [] mChildren;
		}
		mTreeParent = 0;
		mCullBoundingBox = 0;
	}
	//-----------------------------------------------------------------------
	void SpatialTree::reset(void)
	{
		//initialize all children to null.
		for (unsigned int i = 0; i < mTreeManager->getMaxChildrenCount(); i++)
		{
			if (mChildren[i])
			{
				mTreeManager->deleteTree (mChildren[i]);
				mChildren[i] = 0;
			}       
		}
		mTreeParent = 0;  
		mNumNodes = 0;
		mCullBoundingBox = 0;
	};
    //-----------------------------------------------------------------------
	void SpatialTree::setBoundingBox(const Ogre::Vector3 &min, 
		const Ogre::Vector3 &max)
	{
		mBoundingBox.setExtents(min, max);
		setCullBoundingBox(&mBoundingBox);
		if(mWireBoundingBox != 0)
			mWireBoundingBox->setupBoundingBox(mBoundingBox);  
	}			
	//-----------------------------------------------------------------------
	void SpatialTree::setCullBoundingBox(Ogre::AxisAlignedBox *box)
	{
		mCullBoundingBox = box;
		if(hasOcclusionBoundingBox())
			// getOcclusionBoundingBox()->setupBoundingBox(mBoundingBox);
			getOcclusionBoundingBox()->setupBoundingBox(*mCullBoundingBox);
	}
	//-----------------------------------------------------------------------
	Ogre::WireBoundingBox* SpatialTree::getWireBoundingBox()
	{
		// Create a WireBoundingBox if needed
		if (mWireBoundingBox == 0)
		{
			mWireBoundingBox = new Ogre::WireBoundingBox();
			mWireBoundingBox->setupBoundingBox(mBoundingBox);
			//mWireBoundingBox->setupBoundingBox(mCullBoundingBox);
		}

		return mWireBoundingBox;
	}
	//-----------------------------------------------------------------------
	void SpatialTree::_addNode(SpatialTreeNode * node)
	{
		SpatialTreeNode * n = static_cast<SpatialTreeNode *>(node);
		if (n->isStaticNode ())
			mStaticNodes.push_back (n);
		else
			mMovingNodes.push_back (n);
		mNodes.push_back (n);
		n->setTreeParent(this);

		_ref();

	}
	//-----------------------------------------------------------------------
	void SpatialTree::_removeNode(SpatialTreeNode * node)
	{  
		SpatialTreeNode * n = static_cast<SpatialTreeNode *>(node);
		octree_assert (!mNodes.empty());

		SpatialTreeNodeList::iterator it;

		it = std::find (mNodes.begin (), mNodes.end (), n); 
		if (it != mNodes.end ()) 
			mNodes.erase (it);

		if (n->isStaticNode ())
		{
			it = std::find(mStaticNodes.begin (), mStaticNodes.end (), n); 
			if (it != mStaticNodes.end ()) 
				mStaticNodes.erase (it);
		}
		else
		{
			it = std::find(mMovingNodes.begin (), mMovingNodes.end (), n); 
			if (it != mMovingNodes.end ()) 
				mMovingNodes.erase (it);
		}
		n->setTreeParent(0);
		//update total counts.
		if (mNumNodes == 1)
			_unref(this);
		else
			_unref();
	}
	//-----------------------------------------------------------------------
	void SpatialTree::_unref(SpatialTree *oldChild)
	{
		--mNumNodes;
		if (oldChild)
		{
			//remove null children
			SpatialTree *child;

			for (unsigned int i = 0; i < mTreeManager->getMaxChildrenCount(); i++)
			{
				child = mChildren[i];
				if (child && !child->hasChildren())
				{
					mTreeManager->deleteTree(child);
					mChildren[i] = 0;	
				}
			}			
		}
		if (mTreeParent != 0)
		{
			if (mNumNodes == 0)
				mTreeParent->_unref(this);
			else
				mTreeParent->_unref();
		}
	};
	//-----------------------------------------------------------------------
	void SpatialTree::setDebugCorners()
	{
		// use the ratio (OCtreeAAB / thisAAB) / 8
		const Vector3 &treeMax =  mBoundingBox.getMaximum ();
		const Vector3 &treeMin =  mBoundingBox.getMinimum ();

		const Vector3 &sceneMax = mTreeManager->getBoundingBox ().getMaximum ();
		const Vector3 &sceneMin = mTreeManager->getBoundingBox ().getMinimum ();

		const Real width  = sceneMax.x - sceneMin.x;
		const Real height = sceneMax.z - sceneMin.z;
		if (Math::Abs(width) > 1e-08 && Math::Abs(height)  > 1e-08)
		{              
			const Real toScreenDivider = 2.0f / 3;
			const Real screenShift =  - 1.f + toScreenDivider*2;	

			const Real left =	((treeMin.x - sceneMin.x) /  width) * toScreenDivider + screenShift;  
			const Real right =	((treeMax.x - sceneMin.x) /  width) * toScreenDivider + screenShift;  

			const Real bottom = ((treeMin.z - sceneMin.z) /  height) * toScreenDivider + screenShift;      
			const Real top =	((treeMax.z - sceneMin.z) /  height) * toScreenDivider + screenShift;

			getRectangle2d ()->setCorners (left, top, right, bottom);
		}
	} 
}
