/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/
/***************************************************************************
TreeManager.cpp  -  description
-------------------
begin                : Fri Sep 27 2002
copyright            : (C) 2002 by Jon Anderson
email                : janders@users.sf.net
 
Enhancements 2003 - 2004 (C) The OGRE Team
 
***************************************************************************/

#include "OctreeCHCPrecompiledHeaders.h"

#include "OgreSpatialTreeManager.h"

#include "OctreeCHCAxisAlignedBoxSceneQuery.h"
#include "OctreeCHCIntersectionSceneQuery.h"
#include "OctreeCHCPlaneBoundedVolumeListSceneQuery.h"
#include "OctreeCHCRaySceneQuery.h"
#include "OctreeCHCSphereSceneQuery.h"
#include "OgreRenderSystem.h"
#include "OgreWireBoundingBox.h"
#include "OgreRenderOperation.h"
#include "OgreStringConverter.h"

#include "OgreOverlayManager.h"
#include "OgreOverlayElement.h"

#include "OctreeCHCCamera.h"
#include "OctreeCHC.h"
#include "OctreeCHCNode.h"

#include "OctreeCHCOcclusionBoundingBox.h"
#include "OctreeCHCOcclusionSorter.h"
#include "OctreeCHCOcclusionVisibilityData.h"
#include "OctreeCHCOcclusionTraversal.h"
#include "OctreeCHCOcclusionCameraTraversal.h"
#include "OctreeCHCOcclusionVFTraversal.h"
#include "OctreeCHCOcclusionSWTraversal.h"
#include "OctreeCHCOcclusionDebugTraversal.h"
#include "OctreeCHCQueryTraversal.h"


using namespace Ogre;

namespace OgreOctreeCHC
{
    //---------------------------------------------------------------------
    SpatialTreeManager::SpatialTreeManager(OctreeCHCSceneManager *scnMgr, 
		const Ogre::AxisAlignedBox &box, 
		int max_depth,
		const size_t maxChildrenCount) :
		mScnMgr(scnMgr),
		mTree(0),
		mMaxChildrenCount(maxChildrenCount), 
		mMaxDepth(max_depth),
		mBoundingBox(box)
    {
    }
    //---------------------------------------------------------------------
    SpatialTreeManager::~SpatialTreeManager()
    {
		// derivated class is responsible fro deletion
		octree_assert (mTree == 0);
	}
	//---------------------------------------------------------------------
	size_t SpatialTreeManager::getNumNodes() const 
	{
		octree_assert (mTree);
		return mTree->getNumNodes();
	}
    //---------------------------------------------------------------------
    /** Only removes the node from the OctreeCHC.  It leaves the OctreeCHC, even if it's empty.
    */
    void SpatialTreeManager::_removeNode(SpatialTreeNode* node)
	{
		octree_assert (mTree);
        octree_assert (node);
		SpatialTreeNode * tn = static_cast<SpatialTreeNode *> (node);
		SpatialTree * tree = tn->getTreeParent();
		if (tree)
			tree->_removeNode (tn);
		tn->setTreeParent (0);
    }
    //---------------------------------------------------------------------
    void SpatialTreeManager::_insertNode(SpatialTreeNode* n)
    {
		octree_assert (n);
		octree_assert (mTree);
        mTree->_insertNode(n);
	}
	//---------------------------------------------------------------------
	void SpatialTreeManager::traverse(ConstDepthTraversalConst &traverser)
	{
		if (mTree)
			mTree->depthTraversal(traverser);
	}
	//---------------------------------------------------------------------
	void SpatialTreeManager::traverse(DepthTraversalConst &traverser)
	{
		if (mTree)
			mTree->depthTraversal(traverser);
	}
	//---------------------------------------------------------------------
	void SpatialTreeManager::traverse(DepthTraversal &traverser)
	{
		if (mTree)
			mTree->depthTraversal(traverser);
	}
    //---------------------------------------------------------------------
    void SpatialTreeManager::resize(const Ogre::AxisAlignedBox &box, const int depth)
    {
        mMaxDepth = depth;
        resize (box);
    }
    //---------------------------------------------------------------------
    void SpatialTreeManager::resize(const Ogre::AxisAlignedBox &box)
    {
        std::list < SceneNode * > nodes;
		traverse(BoxQueryDepthTraversal(mTree->getBoundingBox (), nodes, 0));

        if (mTree)
            deleteTree (mTree);

        mBoundingBox = box;
        mTree = getNewTree();
        octree_assert (mTree);      
        mTree ->setBoundingBox (box.getMinimum(), box.getMaximum());  
        
        if (!nodes.empty())
        {
            std::list < SceneNode * > ::iterator it;
            it = nodes.begin();

            while (it != nodes.end())
            {
                SpatialTreeNode * const on = static_cast < SpatialTreeNode * > (*it);
                on->setTreeParent(0);
                _updateNode(on);
                ++it;
            } // while (it != nodes.end())
        }  
	}
}

