/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/
/***************************************************************************
TreeNode.cpp  -  description
-------------------
begin                : Fri Sep 27 2002
copyright            : (C) 2002 by Jon Anderson
email                : janders@users.sf.net

Enhancements 2003 - 2004 (C) The OGRE Team

***************************************************************************/

#include "OctreeCHCPrecompiledHeaders.h"

#include "OgreRoot.h"

#include "OgreSpatialTreeNode.h"
#include "OctreeCHCOcclusionBoundingBox.h"
#include "OgreWireBoundingBox.h"
#include "OgreRenderQueue.h"


using namespace Ogre;

namespace OgreOctreeCHC
{   
    //-----------------------------------------------------------------------
    SpatialTreeNode::SpatialTreeNode(Ogre::SceneManager* creator) : 
        SceneNode(creator),
        OcclusionElement(),

        mTreeParent(0),
        mIsStaticNode(true),
        mIsOccluder(true)
    {
       
    }
    //-----------------------------------------------------------------------
    SpatialTreeNode::SpatialTreeNode(Ogre::SceneManager* creator, const Ogre::String& name) : 
        SceneNode(creator, name),
        OcclusionElement(),

        mTreeParent(0),
        mIsStaticNode(true),
        mIsOccluder(true)
    {
        
    }
    //-----------------------------------------------------------------------
    SpatialTreeNode::~SpatialTreeNode()
	{
		removeAllChildren();
		if (mParent)
			mParent->removeChild(this); 
		octree_assert (!mParent);
    }
    //-----------------------------------------------------------------------
    //same as SceneNode, only it doesn't care about children...
    void SpatialTreeNode::_updateBounds(void)
    {
        mWorldAABB.setNull();

        // Update bounds from own attached objects
        ObjectMap::const_iterator i = mObjectsByName.begin();        
        while (i != mObjectsByName.end())
        {
            const MovableObject * const mo = i->second;
            // Get World bounds of object
            mWorldAABB.merge(mo->getWorldBoundingBox(true));

            ++i;
        }

        //update the OctreeCHCSceneManager that things might have moved.
        // if it hasn't been added to the OctreeCHC, add it, and if has moved
        // enough to leave it's current node, we'll update it.

        if (! mWorldAABB.isNull())
        { 
            octree_assert (mCreator);
            static_cast < OctreeCHCSceneManager * > (mCreator)->_updateSceneNode(this);
        }

		//if (hasOcclusionBoundingBox())
		//	getOcclusionBoundingBox()->setupBoundingBox(mWorldAABB);

		setOcclusionBoundingBox(mWorldAABB);
        if (mWireBoundingBox)
            mWireBoundingBox->setupBoundingBox(mWorldAABB);
        if (!mIsRegisteredToCam)
            static_cast< OctreeCHCSceneManager * > (mCreator)->
            registeredNodeInCamera (this);
    }
    //-----------------------------------------------------------------------
     /** Adds the attached objects of this OctreeCHCScene node into the queue. */
	void SpatialTreeNode::_findVisibleObjects
		(
		Camera* cam, 
		RenderQueue* queue, 
		VisibleObjectsBoundsInfo* visibleBounds, 
		bool includeChildren, 
		bool displayNodes, 
		bool onlyShadowCasters
		)
    {

		// Check self visible
		//if (!cam->isVisible(mWorldAABB))
		//	return;

		// Add all entities
		if (!mObjectsByName.empty())
		{
			ObjectMap::iterator iobj;
			ObjectMap::iterator iobjend = mObjectsByName.end();
			for (iobj = mObjectsByName.begin(); iobj != iobjend; ++iobj)
			{
				// Tell attached objects about camera position (in case any extra processing they want to do)
				iobj->second->_notifyCurrentCamera(cam);
				if (iobj->second->isVisible() && 
					(!onlyShadowCasters || iobj->second->getCastShadows()))
				{
					iobj->second->_updateRenderQueue(queue);

					// update visible boundaries AxisAlignedBox
					if (visibleBounds)
					{
						visibleBounds->merge(iobj->second->getWorldBoundingBox(true), 
							iobj->second->getWorldBoundingSphere(true), cam);
					}
				}
			}
		}

		if (includeChildren && !mChildren.empty())
		{
			const ChildNodeMap::iterator childend (mChildren.end());
			for (ChildNodeMap::iterator child = mChildren.begin(); child != childend; ++child)
			{
				SceneNode * const sceneChild = static_cast<SceneNode*> (child->second);
				sceneChild->_findVisibleObjects(cam, 
												queue, 
												visibleBounds, 
												includeChildren, 
												displayNodes, 
												onlyShadowCasters);
			}
		}

		if (displayNodes)
		{
			// Include self in the render queue
			queue->addRenderable(this);
		}

		// Check if the bounding box should be shown.
		// See if our flag is set or if the scene manager flag is set.
		if (mShowBoundingBox || (mCreator && mCreator->getShowBoundingBoxes())) 
		{ 
			_addBoundingBoxToQueue(queue);
		}
    }
    //-----------------------------------------------------------------------
    void SpatialTreeNode::attachObject(Ogre::MovableObject* obj)
    {    
        if (mIsStaticNode)
        {
            const Ogre::String movableType = obj->getMovableType();
            if (movableType == "Entity" || movableType == "StaticGeometry")
            {
                if (StringUtil::startsWith(obj->getName(), "Static", false))
                {
                     if (mIsOccluder && !StringUtil::startsWith(obj->getName(), "StaticOccluder", false))
                     {  
                        mIsOccluder = false;
                     }
                }
                else
                {
                    mIsStaticNode = false;
                    mIsOccluder = false;
                }
            }
            else if (movableType != "TerrainMipMap")
            {
                 mIsStaticNode = false;
                 mIsOccluder = false;
            }
        }
        SceneNode::attachObject (obj);
    }
    //-----------------------------------------------------------------------
    MovableObject* SpatialTreeNode::detachObject(unsigned short index)
    {
		// if we detach 1 object, after the return, 
		// there's none left so we reset the thing.
        if (!mObjectsByName.empty() && 1 == mObjectsByName.size())
        {
            mIsStaticNode = true;
            mIsOccluder = true;
        }
        return SceneNode::detachObject (index);

    }
    //-----------------------------------------------------------------------
    MovableObject* SpatialTreeNode::detachObject(const Ogre::String& name)
    {        
        if (!mObjectsByName.empty() && 1 == mObjectsByName.size())
        {
            mIsStaticNode = true;
            mIsOccluder = true;
        }
        return SceneNode::detachObject (name);
    }
    //-----------------------------------------------------------------------
    void SpatialTreeNode::detachObject(Ogre::MovableObject* obj)
    {
        if (!mObjectsByName.empty() && 1 == mObjectsByName.size())
        {
            mIsStaticNode = true;
            mIsOccluder = true;
        }
        SceneNode::detachObject (obj);
    }
    //-----------------------------------------------------------------------
    void SpatialTreeNode::detachAllObjects(void)
    {
		SceneNode::detachAllObjects ();
        mIsStaticNode = true;
        mIsOccluder = true;
    }
    //-----------------------------------------------------------------------
    void SpatialTreeNode::_addAlreadyNotifiedToVisibles()
    {             
        if (!mNotifiedVisibles.empty())
        {
            OctreeCHCSceneManager * const osm = static_cast < OctreeCHCSceneManager * > (mCreator);
            
            // create the render Queue.
            MovableObjectList::iterator it = mNotifiedVisibles.begin(),
                                        itend = mNotifiedVisibles.end();
            while (it != itend)
			{
				octree_assert ((*it)->isVisible ());
                osm->addVisible (*it);
                ++it;
            }
            mNotifiedVisibles.clear();
        }
    }
    //-----------------------------------------------------------------------
    void SpatialTreeNode::_addToVisibles(Ogre::Camera*  cam, 
                                        const bool onlyShadowCasters)
    {             
        if (!mNotifiedVisibles.empty())
        {
            OctreeCHCSceneManager * const osm = static_cast < OctreeCHCSceneManager * > (mCreator);
            
            // create the render Queue.
           ObjectMap::iterator mit = mObjectsByName.begin ();
            while (mit != mObjectsByName.end ())
            {
                MovableObject * const mo = mit->second;  
                mo->_notifyCurrentCamera (cam); 
                if (mo->isVisible () &&
                    (!onlyShadowCasters || mo->getCastShadows ()))
                {
                    osm->addVisible (mo);
                }
                ++mit;
            }         
            mNotifiedVisibles.clear ();
        }
    }
    //-----------------------------------------------------------------------
    bool SpatialTreeNode::notifyNodeObjects(Ogre::Camera*  cam, 
                                        const bool onlyShadowCasters)
    {       
        // if at least one is visible.
        bool isVisible = false;
        mNotifiedVisibles.clear();
        ObjectMap::iterator mit = mObjectsByName.begin ();
        while (mit != mObjectsByName.end ())
        {
            MovableObject * const mo = mit->second;  
            mo->_notifyCurrentCamera (cam); 
            if (mo->isVisible () &&
                (!onlyShadowCasters || mo->getCastShadows ()))
            {
                mNotifiedVisibles.push_back (mo);
                isVisible = true;
            }
            ++mit;
        }         
        return isVisible;
    }
    //-----------------------------------------------------------------------
    MovableObjectList *SpatialTreeNode::getVisibleNotifiedNodeObjects(Ogre::Camera*  cam, 
                                                                                const bool onlyShadowCasters)
    {            
        notifyNodeObjects(cam, onlyShadowCasters);
        return &mNotifiedVisibles;
    }
   
    //-----------------------------------------------------------------------
    void SpatialTreeNode::addChild(Ogre::Node* child)
    {
        mIsRegisteredToCam = false;
		static_cast< OctreeCHCSceneManager * > (mCreator)->registeredNodeInCamera(this);
        // call superclass method
        Node::addChild (child);
    } 
    //-----------------------------------------------------------------------
    void SpatialTreeNode::_removeNodeAndChildren()
    { 
        octree_assert (mCreator);
		static_cast< OctreeCHCSceneManager * > (mCreator)->unregisteredNodeInCamera (this);
        mIsRegisteredToCam = false;

        SpatialTree * tree = getTreeParent();
        if (tree)
          tree->_removeNode(this);
       setTreeParent (0);

        //remove all the children nodes as well from the OctreeCHC.
        ChildNodeMap::iterator it = mChildren.begin();
        while(it != mChildren.end())
        {
            static_cast<SpatialTreeNode *>(it->second)->_removeNodeAndChildren();
            ++it;
        }
    }
    //-----------------------------------------------------------------------
    void SpatialTreeNode::removeAllChildren(void)
    {
        ChildNodeMap::iterator i, iend = mChildren.end();
        for (i = mChildren.begin(); i != iend; ++i)
        {	
            static_cast <SpatialTreeNode* > (i->second)->_removeNodeAndChildren();	
		}
        Node::removeAllChildren();
    }
    //-----------------------------------------------------------------------
    Node * SpatialTreeNode::removeChild(unsigned short index)
    {
        SpatialTreeNode *on = static_cast<SpatialTreeNode* >(SceneNode::removeChild(index));
        on->_removeNodeAndChildren(); 
        return on; 
    }
    //-----------------------------------------------------------------------
    Node * SpatialTreeNode::removeChild(Ogre::Node* child)
    {
        SpatialTreeNode *on = static_cast<SpatialTreeNode* >(SceneNode::removeChild(child));
        on->_removeNodeAndChildren(); 
        return on; 
    }
    //-----------------------------------------------------------------------
    Node * SpatialTreeNode::removeChild(const Ogre::String & name)
    {
        SpatialTreeNode *on = static_cast< SpatialTreeNode * >(SceneNode::removeChild(name));
        on->_removeNodeAndChildren(); 
        return on; 
    }
    //-----------------------------------------------------------------------
    void SpatialTreeNode::setDebugCorners()
        {
            // use the ratio (OCtreeAAB / thisAAB) 
            const Ogre::Vector3 &sceneMax = static_cast <OctreeCHCSceneManager *> (mCreator)->getBoundingBox ().getMaximum ();
            const Ogre::Vector3 &sceneMin = static_cast <OctreeCHCSceneManager *> (mCreator)->getBoundingBox ().getMinimum ();

			const Real width  = sceneMax.x - sceneMin.x;
			const Real height = sceneMax.z - sceneMin.z;

			const Ogre::Vector3 &max =  mWorldAABB.getMaximum ();
			const Ogre::Vector3 &min =  mWorldAABB.getMinimum ();

			if (Math::Abs(width) > 1e-08 && Math::Abs(height)  > 1e-08)
			{              
				const Real toScreenDivider = 2.0f / 3;
				const Real screenShift =  - 1.f + toScreenDivider*2;
				const Real left =	((min.x - sceneMin.x) /  width) * toScreenDivider + screenShift;  
				const Real bottom = ((min.z - sceneMin.z) /  height) * toScreenDivider + screenShift;
				const Real right =	((max.x - sceneMin.x) /  width) * toScreenDivider + screenShift;        
				const Real top =	((max.z - sceneMin.z) /  height) * toScreenDivider + screenShift;
				getRectangle2d ()->setCorners (left, top, right, bottom);
			}
        }  
       
}
