/*
-----------------------------------------------------------------------------
This source file is part of OGRE
(Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2006 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/
#include "OctreeCHCPrecompiledHeaders.h"

#include "RadarFrustum.h"

using namespace Ogre;

namespace OgreOctreeCHC
{
	//--------------------------------------------------------------------
	RadarFrustum::RadarFrustum()
	{

	}
	//--------------------------------------------------------------------
	void RadarFrustum::SetPerspective( const Radian FOV, const Real ViewAspect, const Real nearZ, const Real farZ )
	{
		//Real degree = FOV.valueDegrees();
		//mRFactor = Math::Tan( FOV  );
		mRFactor = Math::Tan( FOV   );
		//mUFactor = mRFactor * ViewAspect;
		mUFactor = mRFactor * (1.0f / ViewAspect);
		mNearZ   = nearZ;
		mFarZ    = farZ;
	}
	//--------------------------------------------------------------------
	void RadarFrustum::Build(  const Vector3& Eye, const Vector3& Forward, const Vector3& Right, const Vector3& Up )
	{
		mEyePosition   = Eye;
		mForwardVector = Forward;

		// X axis of camera of given "up" vector and Z axis
		//mRightVector = Up.crossProduct(Forward);
		//mRightVector.normalise();
		mRightVector   = Right;

		// the real "up" vector is the cross product of Z and X
		//mUpVector = mRightVector.crossProduct(Forward);
		//mUpVector.normalise();
		mUpVector      = Up;
	}
	//--------------------------------------------------------------------
	bool RadarFrustum::IsPointIn( const Vector3& Point ) const
	{
		const Vector3 OP (Point - mEyePosition);   

		const Real f = OP.dotProduct(mForwardVector);// OP dot ForwardVector 
		if (f < mNearZ || mFarZ < f) 
			return false;

		{
			const Real r = OP.dotProduct(mRightVector); // OP dot RightVector
			const Real rLimit = mRFactor * f;
			if (r < -rLimit || rLimit < r) 
				return false;

		}
		{
			const Real u = OP.dotProduct(mUpVector);// OP dot RightVector
			const Real uLimit = mUFactor * f;
			if (u < -uLimit || uLimit < u) 
				return false;
		}

		// Up to here the point is known to be in the frustum
		return true;
	}
	//--------------------------------------------------------------------
	// This method is highly unoptimized and is presented here just as
	// an example because it is not covered in the article but the algorithm
	// is so similar to the followed in the six-planes approach that it 
	// need no much explanation.
	// please, refer to http://fpuig.cjb.net/ for an optimized method that 
	// use p-n-points and extend transformation.
	bool RadarFrustum::IsAABBIn  ( const Vector3& v1, const Vector3& v2) const
	{
		Vector3 P;
		size_t nOutOfLeft=0, nOutOfRight=0, nOutOfFar=0, nOutOfNear=0, nOutOfTop=0, nOutOfBottom=0;
		bool bIsInRightTest, bIsInUpTest, bIsInFrontTest;

		Vector3 Corners[2];
		Corners[0] = v1 - mEyePosition;
		Corners[1] = v2 - mEyePosition;

		for (size_t i = 0; i < 8; ++i) 
		{
			bIsInRightTest = bIsInUpTest = bIsInFrontTest = false;
			P.x = Corners[      i & 1 ].x;
			P.y = Corners[ (i>>2) & 1 ].y;
			P.z = Corners[ (i>>1) & 1 ].z;

			const Real f = mForwardVector.dotProduct(P);
			{
				if (f < mNearZ) 
					++nOutOfNear;		
				else if (f > mFarZ) 
					++nOutOfFar;
				else 
					bIsInFrontTest = true;
			}

			{
				const Real r = mRightVector.dotProduct(P);
				const Real rLimit = mRFactor*f;


				if (r < -rLimit) 
					++nOutOfLeft;
				else if (r > rLimit) 
					++nOutOfRight;
				else 
					bIsInRightTest = true;
			}

			{
				const Real u = mUpVector.dotProduct(P);
				const Real uLimit = mUFactor*f;

				if (u < -uLimit) 
					++nOutOfBottom;
				else if (u > uLimit) 
					++nOutOfTop;
				else 
					bIsInUpTest = true;

			}

			if (bIsInRightTest && bIsInFrontTest && bIsInUpTest) 
				return true;
		}

		if (nOutOfLeft == 8 || nOutOfRight  == 8 || 
			nOutOfFar  == 8 || nOutOfNear   == 8 || 
			nOutOfTop  == 8 || nOutOfBottom == 8 ) 
			return false;
		return true;
	}
	//--------------------------------------------------------------------
	// This method is highly unoptimized and is presented here just as
	// an example because it is not covered in the article but the algorithm
	// is so similar to the followed in the six-planes approach that it 
	// need no much explanation.
	// please, refer to http://fpuig.cjb.net/ for an optimized method that 
	// use p-n-points and extend transformation.
	bool RadarFrustum::IsOBBIn   ( const Vector3& Center, const Vector3& HalfDimensions, const Quaternion& RotationQuat) const
	{
		Vector3 P;
		int nOutOfLeft=0, nOutOfRight=0, nOutOfFar=0, nOutOfNear=0, nOutOfTop=0, nOutOfBottom=0;
		bool bIsInRightTest, bIsInUpTest, bIsInFrontTest;

		Vector3 Corners[8];
		Corners[0] = Vector3( Center.x+HalfDimensions.x,  Center.y+HalfDimensions.y,  Center.z+HalfDimensions.z);
		Corners[1] = Vector3( Center.x+HalfDimensions.x,  Center.y+HalfDimensions.y,  Center.z-HalfDimensions.z);
		Corners[2] = Vector3( Center.x+HalfDimensions.x,  Center.y-HalfDimensions.y,  Center.z+HalfDimensions.z);
		Corners[3] = Vector3( Center.x+HalfDimensions.x,  Center.y-HalfDimensions.y,  Center.z-HalfDimensions.z);
		Corners[4] = Vector3( Center.x-HalfDimensions.x,  Center.y+HalfDimensions.y,  Center.z+HalfDimensions.z);
		Corners[5] = Vector3( Center.x-HalfDimensions.x,  Center.y+HalfDimensions.y,  Center.z-HalfDimensions.z);
		Corners[6] = Vector3( Center.x-HalfDimensions.x,  Center.y-HalfDimensions.y,  Center.z+HalfDimensions.z);
		Corners[7] = Vector3( Center.x-HalfDimensions.x,  Center.y-HalfDimensions.y,  Center.z-HalfDimensions.z);
		int i;
		for (i=0; i<8; ++i) Corners[i] -= mEyePosition;

		for (i=0; i<8; ++i) {
			bIsInRightTest = bIsInUpTest = bIsInFrontTest = false;
			P=Corners[i];

			Real r= mRightVector.x*P.x + mRightVector.y*P.y + mRightVector.z*P.z;
			Real u= mUpVector.x*P.x + mUpVector.y*P.y + mUpVector.z*P.z;
			Real f= mForwardVector.x*P.x + mForwardVector.y*P.y + mForwardVector.z*P.z;

			if (r< -mRFactor*f) ++nOutOfLeft;
			else if (r > mRFactor*f) ++nOutOfRight;
			else bIsInRightTest=true;

			if (u < -mUFactor*f) ++nOutOfBottom;
			else if (u > mUFactor*f) ++nOutOfTop;
			else bIsInUpTest=true;

			if (f < mNearZ) ++nOutOfNear;		
			else if (f > mFarZ) ++nOutOfFar;
			else bIsInFrontTest=true;

			if (bIsInRightTest && bIsInFrontTest && bIsInUpTest) return true;
		}

		if (nOutOfLeft==8 || nOutOfRight==8 || nOutOfFar==8 || nOutOfNear==8 || nOutOfTop==8 || nOutOfBottom==8 ) return false;
		return true;
	}
	//--------------------------------------------------------------------
	bool RadarFrustum::IsHullIn( const Vector3* aVertices, const char nVertices ) const
	{
		int BehindLeft=0, BehindRight=0, BehindFar=0, BehindNear=0, BehindTop=0, BehindBottom=0;
		bool inForward, inRight, inUp;
		for (char i(0); i<nVertices; ++i)
		{
			inForward=inRight=inUp=false;

			Vector3 OP = aVertices[i] - mEyePosition;   

			Real f = OP.dotProduct(mForwardVector);
			Real r = OP.dotProduct(mRightVector);              // OP dot RightVector
			Real u = OP.dotProduct(mUpVector);              // OP dot RightVector

			Real rLimit = mRFactor * f;
			Real uLimit = mUFactor * f;

			if (f < mNearZ ) ++BehindNear;
			else if (f > mFarZ) ++BehindFar;
			else inForward=true;

			if (r < -rLimit) BehindLeft++;
			else if (r > rLimit) BehindRight++;
			else inRight=true;

			if (u < -uLimit) ++BehindBottom;		
			else if (u > uLimit) ++BehindTop;
			else inUp=true;

			if (inForward && inRight && inUp) return true;
		}

		if (BehindLeft==8 || BehindRight==8 || BehindFar==8 || 
			BehindNear==8 || BehindTop==8 || BehindBottom==8) return false;
		return true;
	}
	//--------------------------------------------------------------------
	bool  RadarFrustum::IsSphereIn( const Vector3& Center, const Real Radius ) const
	{
		const Vector3 OP (Center - mEyePosition);   
		const Real f = OP.dotProduct(mForwardVector);         // OP dot ForwardVector 
		if (f < mNearZ - Radius || mFarZ + Radius < f) 
			return false;

	
		// Unoptimized but more understandable
		{
			const Real r = OP.dotProduct (mRightVector); 
			const Real rLimit = mRFactor * f;
			const Real rTop = rLimit + Radius;
			if (r < -rTop || rTop < r) 
				return false;
		}

		// Optimized ( a substraction is removed )
		{
			const Real u = OP.dotProduct (mUpVector); 
			const Real uLimit = mUFactor * f;
			const Real uTop = uLimit + Radius;
			if (u < -uTop || uTop < u) 
				return false;
		}

		return true;
	}

}

